#include "mainwindow.h"
#include "ui_mainwindow.h"

#include <QModbusRtuSerialMaster>
#include <QSerialPort>
#include <QSerialPortInfo>
#include <QSettings>
#include <QTimer>
#include <QTranslator>

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);

    setConnectionState(ConnectionState::Disconnected);

    resetConnection = false;
    modbusMaster = nullptr;

    // Initialize the timer with the proper connection without launching it.
    timer = new QTimer(this);
    connect(timer, &QTimer::timeout, this, &MainWindow::readAllInputRegisters);

    refreshSerialPortList();
}

MainWindow::~MainWindow()
{
    if (modbusMaster != nullptr) {
        timer->stop();
        modbusMaster->disconnectDevice();
        delete modbusMaster;
    }

    delete ui;
}

//------------------------------------
// Private slots
//------------------------------------

/**
 * Connect button click handler.
 * The "clicked" signal is auto-connected by Qt because of the function name.
 *
 * When the connect button is clicked, we try to establish the Modbus connection
 * on the selected serial port. The CPREG-N2 board uses a serial connection configured
 * at 9600 Bauds with 8 bit data, no parity bit and one stop bit.
 *
 * If the connection fails, the user is warned in the status bar.
 */
void MainWindow::on_connectButton_clicked() {
    if (modbusMaster != nullptr) {
        timer->stop();
        modbusMaster->disconnectDevice();
        delete modbusMaster;
        modbusMaster = nullptr;

        // Activate again the widgets used for the serial port choice
        ui->serialPortListBox->setEnabled(true);
        ui->serialPortListRefreshButton->setEnabled(true);

        ui->connectButton->setText(tr("Connect"));

        setConnectionState(ConnectionState::Disconnected);
    } else {
        setConnectionState(ConnectionState::Connecting);

        modbusMaster = new QModbusRtuSerialMaster(this);
        modbusMaster->setConnectionParameter(QModbusDevice::SerialPortNameParameter, ui->serialPortListBox->currentText());
        modbusMaster->setConnectionParameter(QModbusDevice::SerialParityParameter, QSerialPort::NoParity);
        modbusMaster->setConnectionParameter(QModbusDevice::SerialBaudRateParameter, QSerialPort::Baud9600);
        modbusMaster->setConnectionParameter(QModbusDevice::SerialDataBitsParameter, QSerialPort::Data8);
        modbusMaster->setConnectionParameter(QModbusDevice::SerialStopBitsParameter, QSerialPort::OneStop);
        modbusMaster->setTimeout(1000);
        modbusMaster->setNumberOfRetries(3);

        connect(modbusMaster, &QModbusRtuSerialMaster::stateChanged, this, &MainWindow::modbusStateChanged);

        if (modbusMaster->connectDevice()) {
            // Deactivate the widgets used for the serial port choice
            ui->serialPortListBox->setEnabled(false);
            ui->serialPortListRefreshButton->setEnabled(false);

            ui->connectButton->setText(tr("Disconnect"));
        } else {
            // Error
            delete modbusMaster;
            modbusMaster = nullptr;

            ui->serialPortListBox->setEnabled(true);
            ui->serialPortListRefreshButton->setEnabled(true);
            ui->connectButton->setText(tr("Connect"));

            setConnectionState(ConnectionState::Error);
        }
    }
}

/**
 * Refresh button click handler.
 * The "clicked" signal is auto-connected by Qt because of the function name.
 *
 * When the button is clicked, the serial port list combo box is updated with
 * the found ports.
 */
void MainWindow::on_serialPortListRefreshButton_clicked() {
    refreshSerialPortList();
}

/**
 * This slot is connected to the state changed event of the QModbusDevice used
 * for the Modbus connection.
 *
 * When the serial connection is established, the connection state is set to
 * the "Connected State" and then we perform a first read of the input registers
 * to check if we are really connected to a CPREG-N2 board.
 *
 * @param state New state of the Modbus connection.
 */
void MainWindow::modbusStateChanged(QModbusDevice::State state) {
    switch (state) {
    case QModbusDevice::ConnectedState:
        firstRead = true;
        readAllInputRegisters();
        break;
    default:
        break;
    }
}

/**
 * This slot is connected to the timeout event of the QTimer configured in the
 * constructor of the class.
 *
 * It is used to read periodically the new values of the input registers of
 * the CPREG-N2 board or to reset the Modbus connection in case of a false reply
 * timeout which can occurs with some RS485 USB controllers.
 * This reset cannot be done in the slot connected to the reply finished event
 * because an exception is raised into the QSerial module (at least under
 * Windows 7) when the modbusMaster object is destroyed.
 *
 * If a connection reset is requested after the first reply, we assume there is
 * no CPREG-N2 board connected on the current serial port and we close the
 * connection.
 */
void MainWindow::readAllInputRegisters() {
    if (!resetConnection) {
        QModbusDataUnit request = QModbusDataUnit(QModbusDataUnit::InputRegisters, 0, 1);
        QModbusReply *reply = modbusMaster->sendReadRequest(request, 1);

        if (!reply->isFinished()) {
            connect(reply, &QModbusReply::finished, this, &MainWindow::replyReady);
        } else {
            parseReply(reply);
        }
    } else {
        resetConnection = false;

        timer->stop();
        modbusMaster->disconnectDevice();
        delete modbusMaster;
        modbusMaster = nullptr;

        if (firstRead) {
            // There is no connected CPREG-N2 board or it doesn't reply the proper data
            ui->serialPortListBox->setEnabled(true);
            ui->serialPortListRefreshButton->setEnabled(true);
            ui->connectButton->setText(tr("Connect"));

            setConnectionState(ConnectionState::Error);
        } else {
            // Performing the new connection
            modbusMaster = new QModbusRtuSerialMaster(this);
            modbusMaster->setConnectionParameter(QModbusDevice::SerialPortNameParameter, ui->serialPortListBox->currentText());
            modbusMaster->setConnectionParameter(QModbusDevice::SerialParityParameter, QSerialPort::NoParity);
            modbusMaster->setConnectionParameter(QModbusDevice::SerialBaudRateParameter, QSerialPort::Baud9600);
            modbusMaster->setConnectionParameter(QModbusDevice::SerialDataBitsParameter, QSerialPort::Data8);
            modbusMaster->setConnectionParameter(QModbusDevice::SerialStopBitsParameter, QSerialPort::OneStop);
            modbusMaster->setTimeout(1000);
            modbusMaster->setNumberOfRetries(3);

            connect(modbusMaster, &QModbusRtuSerialMaster::stateChanged, this, &MainWindow::modbusStateChanged);

            if (!modbusMaster->connectDevice()) {
                delete modbusMaster;
                modbusMaster = nullptr;

                ui->serialPortListBox->setEnabled(true);
                ui->serialPortListRefreshButton->setEnabled(true);
                ui->connectButton->setText(tr("Connect"));

                setConnectionState(ConnectionState::TimeOut);
            }
        }
    }
}

/**
 * This slot is connected to the finished event of the QModbusReply used
 * to perform the input register read.
 */
void MainWindow::replyReady() {
    auto reply = qobject_cast<QModbusReply *>(sender());

    if (!reply) {
        // Error -> reset the connection
        resetConnection = true;
    } else {
        parseReply(reply);
    }
}

//------------------------------------
// Private methods
//------------------------------------

/**
 * Refreshes the serial port list in the dedicated combo box.
 */
void MainWindow::refreshSerialPortList() {
    const auto infos = QSerialPortInfo::availablePorts();

    ui->serialPortListBox->clear();
    for (const QSerialPortInfo &info : infos) {
        ui->serialPortListBox->addItem(info.portName());
    }
}

/**
 * Refreshes the status bar text according to the current connection
 * state.
 */
void MainWindow::refreshStatusBar() {
    switch (connectionState) {
    case ConnectionState::Connected:
        ui->statusLabel->setStyleSheet("color: black");
        ui->statusLabel->setText(tr("Connected."));
        break;
    case ConnectionState::Disconnected:
        ui->statusLabel->setStyleSheet("color: black");
        ui->statusLabel->setText(tr("Disconnected."));
        break;
    case ConnectionState::Connecting:
        ui->statusLabel->setStyleSheet("color: black");
        ui->statusLabel->setText(tr("Connecting..."));
        break;
    case ConnectionState::TimeOut:
        ui->statusLabel->setStyleSheet("color: red");
        ui->statusLabel->setText(tr("Connection timed out."));
        break;
    case ConnectionState::Error:
        ui->statusLabel->setStyleSheet("color: red");
        ui->statusLabel->setText(tr("Connection failed."));
        break;
    default:
        break;
    }
}

/**
 * Sets the connection state and refresh the status bar.
 *
 * @param state Connection state to set.
 */
void MainWindow::setConnectionState(ConnectionState state) {
    connectionState = state;
    refreshStatusBar();
}

/**
 * Parses the data stored into the given Modbus reply.
 *
 * @param reply Modbus reply to parse.
 */
void MainWindow::parseReply(QModbusReply *reply) {
    QModbusDevice::Error error = reply->error();

    if (error == QModbusDevice::NoError) {
        const QModbusDataUnit result = reply->result();
        uint count = result.valueCount();

        if (count == 1) {
            if (firstRead) {
                firstRead = false;

                // We can now consider the connection as correctly established
                setConnectionState(ConnectionState::Connected);

                // Start timer to read the input registers each 500ms
                timer->setSingleShot(false);
                timer->start(500);
            }

            ui->inputRegister0->setText(QString("%1").arg(result.value(0), 2, 16, QChar('0')).toUpper());
        } else {
            // Error -> reset the connection
            resetConnection = true;
            timer->setSingleShot(true);
            timer->start(100);
        }
    } else {
        // Error -> reset the connection
        resetConnection = true;
        timer->setSingleShot(true);
        timer->start(100);
    }
}
